/******************************************************************************
* This is based on Adafruits HID keyboard example for the nRF52 based
* Bluefruit LE modules, and implements a BLE footswitch controller based on the
* BOSS FS-6 pedal, or any other active low momentary foot switches
*
* Adafruit invests time and resources providing open source code so I want to
* ask that you please support Adafruit and open-source hardware by purchasing
* products from Adafruit!
*
* MIT License
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
* V1_0: initial version, one pedal (two foot switches)
* V1_1: Dual pedal inputs (4 foot switches total) and active buzzer
* V1_2: Flipped AB and CD input jacks, fixed bug in mode shift where release
*       of pedals had to happen during blinking pattern or it would go to next
*       mode
* V1_3: Added multimedia control mode. Added persistent storage of mode.
*       Increased BLE Tx power to 8dBm
***********************************************************************/

#include <bluefruit.h>

// Set up file system for persistent storage of mode variable in flash
#include <Adafruit_LittleFS.h>
#include <InternalFileSystem.h>
#define MODE_FILENAME "/mode.txt"  //File in internal file system to store mode
using namespace Adafruit_LittleFS_Namespace;
File file(InternalFS);

// Pin definitions
#define BUZZER_PIN      12
#define SWITCH_A_PIN    10
#define SWITCH_B_PIN    11
#define SWITCH_C_PIN    7
#define SWITCH_D_PIN    9

// Constant definitions
#define SWITCHDEBOUNCE  50
#define SWITCH_NONE     0
#define SWITCH_A        1
#define SWITCH_B        2
#define SWITCH_C        3
#define SWITCH_D        4
// Multimedia key definitions
#define SCAN_NEXT_TRACK       0xB5
#define SCAN_PREVIOUS_TRACK   0xB6
#define STOP                  0xB7
#define PLAY_PAUSE            0xCD

BLEDis bledis;
BLEHidAdafruit blehid;

byte mode = 0;                      // Keep track of current mode
byte currentSwitch;                 // Currently pressed switch
bool A_switch;                      // Boss FS-6 footswitch 1 A
bool B_switch;                      // Boss FS-6 footswitch 1 B
bool C_switch;                      // Boss FS-6 footswitch 2 A
bool D_switch;                      // Boss FS-6 footswitch 2 B
bool previous_A_switch = true;      // Variable to keep last state
bool previous_B_switch = true;      // Variable to keep last state
bool previous_C_switch = true;      // Variable to keep last state
bool previous_D_switch = true;      // Variable to keep last state
long switch_timestamp = millis();   // Timestamp to handle debouncing

/*******************************************************************************
* Keyboard codes and modifiers for the different modes for A,B,C and D
* Mode 0 page turner: A=page up,B=page dwn, C=home, D=end
* Mode 1 Garageband:  A=Record, B=Start/stop, C=Unused, D= to beginning
* Mode 2 Ableton:     A=Record, B=Play from start/stop, C= Continue play, D=to beginning
* Mode 3 Multimedia:  A
*******************************************************************************/
uint8_t keyPresses[3][4] =  // Keypress codes
{
  { HID_KEY_PAGE_UP,   HID_KEY_PAGE_DOWN,HID_KEY_HOME,   HID_KEY_END},       // Mode 0 (Page Turner)
  { HID_KEY_R,         HID_KEY_SPACE,    HID_KEY_NONE,   HID_KEY_RETURN},    // Mode 1 (Garage Band)
  { HID_KEY_F9,        HID_KEY_SPACE,    HID_KEY_SPACE,  HID_KEY_HOME},      // Mode 2 (Ableton Live)
};
uint8_t modifiers[3][4] =   // Keyboard modifiers
{
  { HID_KEY_NONE,      HID_KEY_NONE,     HID_KEY_NONE,       HID_KEY_NONE},  // Mode 0 (Page turner)
  { HID_KEY_NONE,      HID_KEY_NONE,     HID_KEY_NONE,       HID_KEY_NONE},  // Mode 1 (Garage Band)
  { HID_KEY_NONE,      HID_KEY_NONE,     HID_KEY_SHIFT_LEFT, HID_KEY_NONE},  // Mode 2 (Ableton Live)
};
int16_t Multimediacontrol[4] =  // Keypress codes for multimedia mode
{ PLAY_PAUSE,        STOP,             SCAN_PREVIOUS_TRACK, SCAN_NEXT_TRACK}; // Mode 3 (Multimedia)



void setup()
{
  Serial.begin(115200);
  pinMode(SWITCH_A_PIN,INPUT_PULLUP);     // Set up button inputs with pullup
  pinMode(SWITCH_B_PIN,INPUT_PULLUP);     // Set up button inputs with pullup
  pinMode(SWITCH_C_PIN,INPUT_PULLUP);     // Set up button inputs with pullup
  pinMode(SWITCH_D_PIN,INPUT_PULLUP);     // Set up button inputs with pullup
  pinMode(BUZZER_PIN,OUTPUT);             // Set up output for buzzer
  digitalWrite(BUZZER_PIN, LOW);          // Set up output for buzzer
  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, LOW);         // Set LED OFF

  Bluefruit.begin();
  Bluefruit.setTxPower(8);    // Check bluefruit.h for supported values
  Bluefruit.setName("nRFpedal");
  // Configure and Start Device Information Service
  bledis.setManufacturer("Adafruit");
  bledis.setModel("Bluefruit52");
  bledis.begin();

  /* Start BLE HID
   * Note: Apple requires BLE device must have min connection interval >= 20m
   * ( The smaller the connection interval the faster we could send data).
   * However for HID and MIDI device, Apple could accept min connection interval
   * up to 11.25 ms. Therefore BLEHidAdafruit::begin() will try to set the min and max
   * connection interval to 11.25  ms and 15 ms respectively for best performance.
   */
  blehid.begin();

  // Set callback for set LED from central
  blehid.setKeyboardLedCallback(set_keyboard_led);

  /* Set connection interval (min, max) to your perferred value.
   * Note: It is already set by BLEHidAdafruit::begin() to 11.25ms - 15ms
   * min = 9*1.25=11.25 ms, max = 12*1.25= 15 ms
   */
  /* Bluefruit.Periph.setConnInterval(9, 12); */

  // Set up and start advertising
  startAdv();

  // start the file system lib and see if we have anything persisted
  InternalFS.begin();
  // InternalFS.format(); // Uncomment to wipe out all persisted data

  // read stored mode
  file.open(MODE_FILENAME, FILE_O_READ);  // Open file for reading
  if (file)
  {
    uint32_t readLen;
    char buffer[2] = { 0 };     // buffer starts as an empty char array (C string)
    readLen = file.read(buffer, sizeof(buffer));
    buffer[readLen] = 0;        // drop any last character and make sure the buffer contains a C string
    mode = atoi(buffer);        // convert the string in the file to a number
    Serial.println(mode);       // Print the mode read from flash
    file.close();
  }
  else
  {
    Serial.println("Mode file does not yet exist, will be created at next mode change.");
  }
  for (byte i = 0; i<(mode+1); i++)     // Indicate with buzzer beeps and LED blinks the current mode
  {
    digitalWrite(LED_BUILTIN, HIGH);    // Blink LED...
    digitalWrite(BUZZER_PIN, HIGH);     // And beep...
    delay(300);
    digitalWrite(BUZZER_PIN, LOW);
    digitalWrite(LED_BUILTIN, LOW);
    delay(200);
  }

}

void startAdv(void)
{
  // Advertising packet
  Bluefruit.Advertising.addFlags(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);
  Bluefruit.Advertising.addTxPower();
  Bluefruit.Advertising.addAppearance(BLE_APPEARANCE_HID_KEYBOARD);

  // Include BLE HID service
  Bluefruit.Advertising.addService(blehid);

  // There is enough room for the dev name in the advertising packet
  Bluefruit.Advertising.addName();

  /* Start Advertising
   * - Enable auto advertising if disconnected
   * - Interval:  fast mode = 20 ms, slow mode = 152.5 ms
   * - Timeout for fast mode is 30 seconds
   * - Start(timeout) with timeout = 0 will advertise forever (until connected)
   *
   * For recommended advertising interval
   * https://developer.apple.com/library/content/qa/qa1931/_index.htmlMMMMMNNNN
   */
  Bluefruit.Advertising.restartOnDisconnect(true);
  Bluefruit.Advertising.setInterval(32, 244);    // in unit of 0.625 ms
  Bluefruit.Advertising.setFastTimeout(30);      // number of seconds in fast mode
  Bluefruit.Advertising.start(0);                // 0 = Don't stop advertising after n seconds
}

void loop()
{
  A_switch = digitalRead(SWITCH_A_PIN);  // Read current state of the pin
  B_switch = digitalRead(SWITCH_B_PIN);  // Read current state of the pin
  C_switch = digitalRead(SWITCH_C_PIN);  // Read current state of the pin
  D_switch = digitalRead(SWITCH_D_PIN);  // Read current state of the pin
  if ((A_switch == false) && (B_switch == false))  //IF both A and B pressed -> Mode switch
  {
    long timer = millis();                    //Start to measure time held down
    while ((A_switch == false) && (B_switch == false))  //As long as both A and B pressed...
    {
      if (millis() - timer > 3000)            //If held for more than 3 seconds
      {
        mode = mode + 1;
        if (mode > 3)
        {
          mode = 0;
        }
        for (byte i = 0; i<(mode+1); i++)     // Indicate with buzzer beeps and LED blinks the current mode
        {
          digitalWrite(LED_BUILTIN, HIGH);
          digitalWrite(BUZZER_PIN, HIGH);
          delay(300);
          digitalWrite(BUZZER_PIN, LOW);
          digitalWrite(LED_BUILTIN, LOW);
          delay(200);
        }
        timer = millis();                      // restart timer
      }
      A_switch = digitalRead(SWITCH_A_PIN);  // Read current state of the pin
      B_switch = digitalRead(SWITCH_B_PIN);  // Read current state of the pin
    }
    if(file.open(MODE_FILENAME, FILE_O_WRITE)) // Write the new mode to the filesystem
    {
      char buffer[2] = {0};
      itoa(mode, buffer, 10);
      file.seek(0);                           // Move pointer to start of file (write automatically appends file)
      file.write(buffer, strlen(buffer));     // Write mode variable as ASCII to mode.txt
      file.close();
      Serial.print("New mode: ");
      Serial.println(mode);
    }
  }
  else if ((A_switch == false) && (previous_A_switch == true) && ((millis() - switch_timestamp) > SWITCHDEBOUNCE)) //If button has changed state and more than "SWITCHDEBOUNCE" ms since last press
  {
    currentSwitch = SWITCH_A;
    Serial.print("A ");
  }
  else if ((B_switch == false) && (previous_B_switch == true) && ((millis() - switch_timestamp) > SWITCHDEBOUNCE)) //If button has changed state and more than "SWITCHDEBOUNCE" ms since last press
  {
    currentSwitch = SWITCH_B;
    Serial.print("B ");
  }
  else if ((C_switch == false) && (previous_C_switch == true) && ((millis() - switch_timestamp) > SWITCHDEBOUNCE)) //If button has changed state and more than "SWITCHDEBOUNCE" ms since last press
  {
    currentSwitch = SWITCH_C;
    Serial.print("C ");
  }
  else if ((D_switch == false) && (previous_D_switch == true) && ((millis() - switch_timestamp) > SWITCHDEBOUNCE)) //If button has changed state and more than "SWITCHDEBOUNCE" ms since last press
  {
    currentSwitch = SWITCH_D;
    Serial.print("D ");
  }
  else
  {
    currentSwitch = SWITCH_NONE; //Indicate no switch pressed
  }

  if (currentSwitch)            // If a switch was pressed
  {
    if (mode == 3)  //If Multimedia mode
    {
      blehid.consumerKeyPress(Multimediacontrol[currentSwitch-1]);
      Serial.println(Multimediacontrol[currentSwitch-1],HEX);
      blehid.consumerKeyRelease();
    }
    else
    {
      uint8_t keycodes[6] = { keyPresses[mode][currentSwitch-1], HID_KEY_NONE , HID_KEY_NONE , HID_KEY_NONE , HID_KEY_NONE , HID_KEY_NONE };
      blehid.keyboardReport(modifiers[mode][currentSwitch-1], keycodes);
      Serial.print(modifiers[mode][currentSwitch-1],HEX);
      Serial.print(" ");
      Serial.println(keyPresses[mode][currentSwitch-1],HEX);
      blehid.keyRelease();
    }
    switch_timestamp = millis();       // Reset timestamp for when toggle happened (for debouncing logic)
  }
  previous_A_switch = A_switch;
  previous_B_switch = B_switch;
  previous_C_switch = C_switch;
  previous_D_switch = D_switch;
}

/**
 * Callback invoked when received Set LED from central.
 * Must be set previously with setKeyboardLedCallback()
 *
 * The LED bit map is as follows: (also defined by KEYBOARD_LED_* )
 *    Kana (4) | Compose (3) | ScrollLock (2) | CapsLock (1) | Numlock (0)
 */
void set_keyboard_led(uint16_t conn_handle, uint8_t led_bitmap)
{
  (void) conn_handle;

  Serial.print("Keyboard_LED_handler ");
  Serial.println(led_bitmap,BIN);
}
